#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
    Filters for generating python. All filters in this
    module will be available in the template's global namespace as ``py``.
"""
import typing
import keyword
import builtins
import pydsdl

from .. import SupportsTemplateEnv, templateEnvironmentFilter
from .c import VariableNameEncoder
from . import _UniqueNameGenerator


@templateEnvironmentFilter
def filter_to_template_unique_name(env: SupportsTemplateEnv, base_token: str) -> str:
    """
    Filter that takes a base token and forms a name that is very
    likely to be unique within the template the filter is invoked. This
    name is also very likely to be a valid Python identifier.

    .. IMPORTANT::

        The exact tokens generated may change between major or minor versions
        of this library. The only guarantee provided is that the tokens
        will be stable for the same version of this library given the same
        input.

        Also note that name uniqueness is only likely within a given template.
        Between templates there is no guarantee of uniqueness and,
        since this library does not lex generated source, there is no guarantee
        that the generated name does not conflict with a name generated by
        another means.

    .. invisible-code-block: python

        from nunavut.lang.py import filter_to_template_unique_name

    .. code-block:: python

        # Given
        template  = '{{ "f" | to_template_unique_name }},{{ "f" | to_template_unique_name }},'
        template += '{{ "f" | to_template_unique_name }},{{ "bar" | to_template_unique_name }}'

        # then
        rendered = '_f0_,_f1_,_f2_,_bar0_'

    .. invisible-code-block: python

        jinja_filter_tester(filter_to_template_unique_name, template, rendered)

    .. code-block:: python

        # Given
        template = '{{ "i like coffee" | to_template_unique_name }}'

        # then
        rendered = '_i like coffee0_'

    .. invisible-code-block: python

        jinja_filter_tester(filter_to_template_unique_name, template, rendered)

    :param str base_token: A token to include in the base name.
    :returns: A name that is likely to be valid python identifier and is likely to
        be unique within the file generated by the current template.
    """
    return _UniqueNameGenerator.ensure_generator_in_globals(env.globals)('py', base_token, '_', '_')


PYTHON_RESERVED_IDENTIFIERS = frozenset(map(str, list(keyword.kwlist) + dir(builtins)))  # type: typing.FrozenSet[str]


def filter_id(instance: typing.Any, stropping_suffix: str = '_', encoding_prefix: str = 'ZX') -> str:
    """
    Filter that produces a valid Python identifier for a given object. The encoding may not
    be reversable.

    .. invisible-code-block: python

        from nunavut.lang.py import filter_id

    .. code-block:: python

        # Given
        I = 'I like python'

        # and
        template = '{{ I | id }}'

        # then
        rendered = 'I_like_python'


    .. invisible-code-block: python

        jinja_filter_tester(filter_id, template, rendered, I=I)

    .. code-block:: python

        # Given
        I = '&because'

        # and
        template = '{{ I | id }}'

        # then
        rendered = 'ZX0026because'


    .. invisible-code-block: python

        jinja_filter_tester(filter_id, template, rendered, I=I)


    .. code-block:: python

        # Given
        I = 'if'

        # and
        template = '{{ I | id }}'

        # then
        rendered = 'if_'


    .. invisible-code-block: python

        jinja_filter_tester(filter_id, template, rendered, I=I)


    .. code-block:: python

        # Given
        I = 'if'

        # and
        template = '{{ I | id("stropped_") }}'

        # then
        rendered = 'ifstropped_'


    .. invisible-code-block: python

        jinja_filter_tester(filter_id, template, rendered, I=I)


    :param any instance:        Any object or data that either has a name property or can be converted
                                to a string.
    :param str stropping_suffix: String appended to the resolved instance name if the encoded value
                                is a reserved keyword in python.
    :param str encoding_prefix: The string to insert before any four digit unicode number used to represent
                                an illegal character.
                                Note that the caller must ensure the prefix itself consists of only valid
                                characters for Python identifiers.
    :returns: A token that is a valid Python identifier, is not a reserved keyword, and is transformed
              in a deterministic manner based on the provided instance.
    """
    if hasattr(instance, 'name'):
        raw_name = str(instance.name)  # type: str
    else:
        raw_name = str(instance)

    # We use the C variable name encoder since the variable token rules are
    # compatible.
    return VariableNameEncoder('', stropping_suffix, encoding_prefix, False).strop(raw_name,
                                                                                   PYTHON_RESERVED_IDENTIFIERS)


def filter_full_reference_name(t: pydsdl.CompositeType, stropping: bool = True) -> str:
    """
    Provides a string that is the full namespace, typename, major, and minor version for a given composite type.

    .. invisible-code-block: python

        from nunavut.lang.py import filter_full_reference_name

        dummy = lambda: None
        dummy_version = lambda: None
        setattr(dummy, 'version', dummy_version)

    .. code-block:: python

        # Given
        full_name = 'any.str.2Foo'
        major = 1
        minor = 2

        # and
        template = '{{ my_obj | full_reference_name }}'

        # then
        rendered = 'any_.str_._2Foo_1_2'

    .. invisible-code-block: python


        setattr(dummy_version, 'major', major)
        setattr(dummy_version, 'minor', minor)
        setattr(dummy, 'full_name', full_name)
        setattr(dummy, 'short_name', full_name.split('.')[-1])
        jinja_filter_tester(filter_full_reference_name, template, rendered, my_obj=dummy)


    .. code-block:: python

        # Given
        full_name = 'any.str.2Foo'
        major = 1
        minor = 2

        # and
        template = '{{ my_obj | full_reference_name(stropping=False) }}'

        # then
        rendered = 'any.str.2Foo_1_2'

    .. invisible-code-block: python


        setattr(dummy_version, 'major', major)
        setattr(dummy_version, 'minor', minor)
        setattr(dummy, 'full_name', full_name)
        setattr(dummy, 'short_name', full_name.split('.')[-1])
        jinja_filter_tester(filter_full_reference_name, template, rendered, my_obj=dummy)

    :param pydsdl.CompositeType t: The DSDL type to get the fully-resolved reference name for.
    :param bool stropping: If True then the :func:`filter_id` filter is applied to each component in the identifier.
    """
    ns_parts = t.full_name.split('.')
    if len(ns_parts) > 1:
        if stropping:
            ns = list(map(filter_id, ns_parts[:-1]))
        else:
            ns = ns_parts[:-1]

    return '.'.join(ns + [filter_short_reference_name(t, stropping)])


def filter_short_reference_name(t: pydsdl.CompositeType, stropping: bool = True) -> str:
    """
    Provides a string that is a shorted version of the full reference name. This type is unique only within its
    namespace.

     .. invisible-code-block: python

        from nunavut.lang.py import filter_short_reference_name

        dummy = lambda: None
        dummy_version = lambda: None
        setattr(dummy, 'version', dummy_version)

    .. code-block:: python

        # Given
        short_name = '2Foo'
        major = 1
        minor = 2

        # and
        template = '{{ my_obj | short_reference_name }}'

        # then ('str' is stropped to 'str_' before the version is suffixed)
        rendered = '_2Foo_1_2'

    .. invisible-code-block: python

        setattr(dummy_version, 'major', major)
        setattr(dummy_version, 'minor', minor)
        setattr(dummy, 'short_name', short_name)
        jinja_filter_tester(filter_short_reference_name, template, rendered, my_obj=dummy)


    .. code-block:: python

        # Given
        short_name = '2Foo'
        major = 1
        minor = 2

        # and
        template = '{{ my_obj | short_reference_name(stropping=False) }}'

        # then
        rendered = '2Foo_1_2'

    .. invisible-code-block: python

        setattr(dummy_version, 'major', major)
        setattr(dummy_version, 'minor', minor)
        setattr(dummy, 'short_name', short_name)
        jinja_filter_tester(filter_short_reference_name, template, rendered, my_obj=dummy)

    :param pydsdl.CompositeType t: The DSDL type to get the reference name for.
    """
    short_name = '{short}_{major}_{minor}'.format(short=t.short_name, major=t.version.major, minor=t.version.minor)
    if stropping:
        return filter_id(short_name)
    else:
        return short_name


def filter_alignment_prefix(offset: pydsdl.BitLengthSet) -> str:
    """
    Provides a string prefix based on a given :class:`pydsdl.BitLengthSet`.

    .. invisible-code-block: python

        from nunavut.lang.py import filter_alignment_prefix
        import pydsdl

    .. code-block:: python

        # Given
        B = pydsdl.BitLengthSet(32)

        # and
        template = '{{ B | alignment_prefix }}'

        # then ('str' is stropped to 'str_' before the version is suffixed)
        rendered = 'aligned'

    .. invisible-code-block: python

        jinja_filter_tester(filter_alignment_prefix, template, rendered, B=B)


    .. code-block:: python

        # Given
        B = pydsdl.BitLengthSet(32)
        B.increment(1)

        # and
        template = '{{ B | alignment_prefix }}'

        # then ('str' is stropped to 'str_' before the version is suffixed)
        rendered = 'unaligned'

    .. invisible-code-block: python

        jinja_filter_tester(filter_alignment_prefix, template, rendered, B=B)


    :param pydsdl.BitLengthSet offset: A bit length set to test for alignment.
    :return: 'aligned' or 'unaligned' based on the state of the ``offset`` argument.
    """
    if isinstance(offset, pydsdl.BitLengthSet):
        return 'aligned' if offset.is_aligned_at_byte() else 'unaligned'
    else:  # pragma: no cover
        raise TypeError('Expected BitLengthSet, got {}'.format(type(offset).__name__))


def filter_imports(t: pydsdl.CompositeType) -> typing.List[str]:
    """
    Returns a list of all modules that must be imported to used a given type.

    :param pydsdl.CompositeType t: The type to scan for dependencies.
    :return: a list of python module names the provided type depends on.
    """
    # Make a list of all attributes defined by this type
    if isinstance(t, pydsdl.ServiceType):
        atr = t.request_type.attributes + t.response_type.attributes
    else:
        atr = t.attributes

    # Extract data types of said attributes; for type constructors such as arrays extract the element type
    dep_types = list(map(lambda x: x.data_type, atr))  # type: ignore
    for t in dep_types[:]:
        if isinstance(t, pydsdl.ArrayType):
            dep_types.append(t.element_type)

    # Make a list of unique full namespaces of referenced composites
    return list(sorted(set(x.full_namespace for x in dep_types if isinstance(x, pydsdl.CompositeType))))


def filter_longest_id_length(attributes: typing.List[pydsdl.Attribute], stropping: bool = True) -> int:
    """
    Return the length of the longest identifier in a list of :class:`pydsdl.Attribute` objects.

    .. invisible-code-block: python

        from nunavut.lang.py import filter_longest_id_length


    .. code-block:: python

        # Given
        I = ['one.str.int.any', 'three.str.int.any']

        # and
        template = '{{ I | longest_id_length }}'

        # then
        rendered = '32'

    .. invisible-code-block: python

        jinja_filter_tester(filter_longest_id_length, template, rendered, I=I)


    .. code-block:: python

        # Given
        I = ['one.str.int.any', 'three.str.int.any']

        # and
        template = '{{ I | longest_id_length(stropping=False) }}'

        # then
        rendered = '17'

    .. invisible-code-block: python

        jinja_filter_tester(filter_longest_id_length, template, rendered, I=I)


    :param bool stropping: If True then the :func:`filter_id` filter is applied to each attribute identifier
                           before comparing lengths else the attributes are not modified.
    """
    if stropping:
        return max(map(len, map(filter_id, attributes)))
    else:
        return max(map(len, attributes))


def filter_bit_length_set(values: typing.Optional[typing.Union[typing.Iterable[int], int]]) -> pydsdl.BitLengthSet:
    """
    Convert an integer or a list of integers into a :class:`pydsdl.BitLengthSet`.

    .. invisible-code-block: python

        from nunavut.lang.py import filter_bit_length_set
        import pydsdl

        assert type(filter_bit_length_set(23)) == pydsdl.BitLengthSet

    """
    return pydsdl.BitLengthSet(values)
